/* Copyright (C) 2014 2015 Jason Self <j@jxself.org>

This file is part of skeinsum.

skeinsum is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by 
the Free Software Foundation, either version 3 of the License, or 
(at your option) any later version.

skeinsum is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with skeinsum. If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <getopt.h>
#include <malloc.h>
#include <math.h>
#include <glob.h>
#include <sys/stat.h>
#include <errno.h>
#include "SHA3api_ref.h"

#define WARN(msg, ...) fprintf(stderr, "skein%dsum: " msg, hashbitlen, ##__VA_ARGS__)

#define TRYHELP_GOODBYE() do { printf("Try 'skein%dsum --help' for more information.\n", hashbitlen); exit(1); } while(0)

typedef long long unsigned LLU;

extern const int hashbitlen;

#define skeinVersion "1.3"

const size_t input_minbufsize = 32 * 1024;
const size_t input_maxbufsize = 32 * 1024 * 1024;


enum
{
  QUIET_OPTION = 11,
  STATUS_OPTION,
  TAG_OPTION
};

static struct option const long_options[] =
{
  { "binary",  no_argument, NULL, 'b' },
  { "check",   no_argument, NULL, 'c' },
  { "quiet",   no_argument, NULL, QUIET_OPTION },
  { "status",  no_argument, NULL, STATUS_OPTION },
  { "text",    no_argument, NULL, 't' },
  { "warn",    no_argument, NULL, 'w' },
  { "tag",     no_argument, NULL, TAG_OPTION },
  { "help",    no_argument, NULL, 'h' },
  { "version", no_argument, NULL, 'V' },
  { NULL, 0, NULL, 0 }
};

void hash2hexstr(unsigned char hash[], char str[])
{
	int i;
	for (i = 0; i < hashbitlen / 8; i++) {
		sprintf(&str[i * 2], "%02X", hash[i]);
	}
	str[i * 2 + 1] = '\0';
}

int HashFile(const char file_name[], char MsgDigest[], char mode)
{
	int is_stdin = (strcmp(file_name, "-") == 0);

	/* Try to get file info */
	struct stat st;
	st.st_size = 0;  /* ..needed when reading from stdio */
	if (!is_stdin && stat(file_name, &st) < 0) {
		WARN("%s: cannot stat: %s\n", file_name, strerror(errno));
		return -1;
	}

	/* Get filesize */
	size_t fsize = st.st_size;
	if (fsize != st.st_size) {
		WARN("%s: SIZE WARNING: filesize %llu is too big for reading into memory!\n",
			file_name, (long long unsigned)st.st_size);
	}

	/* Open file */
	FILE *fp_in = is_stdin ? stdin : fopen(file_name, (mode == 't' ? "r" : "rb") );
	if (!fp_in) {
		WARN("%s: cannot open: %s\n", file_name, strerror(errno));
		return -1;
	}

	/* Allocate buffer */
	size_t bufsize = fsize + 1;
	if (bufsize < input_minbufsize)
		bufsize = input_minbufsize;

	char *readbuf = malloc (bufsize);
	if (!readbuf && bufsize > input_maxbufsize) {  /* ..Try to get contents by smaller portions */
		bufsize = input_maxbufsize;
		readbuf = malloc (bufsize);
	}

	if (!readbuf) {
		WARN("%s: MEM FAILED: error %s\n", file_name, strerror(errno));
		if (!is_stdin)
			fclose(fp_in);
		return -1;
	}

	/* Read contents */
	size_t readpos = 0, total_readed = 0;

	while (1) {
		size_t maxread = bufsize - readpos;
		size_t readed = fread(readbuf + readpos, 1, maxread, fp_in);
		if (readed > 0 && readed <= maxread)
			total_readed += readed;
		if (readed != maxread)
			break;
		if (getenv("SKEIN_DEBUG"))
			printf("DEBUG: bufsize=%llu (0x%llx), readpos=%llu (0x%llx), maxread=%llu (0x%llx), total=%llu (0x%llx)\n",
				(LLU)bufsize, (LLU)bufsize, (LLU)readpos, (LLU)readpos,
				(LLU)maxread, (LLU)maxread, (LLU)total_readed, (LLU)total_readed);
		char *newbuf = NULL;
		if (bufsize * 2 > bufsize)  /* ..check overflow */
			newbuf = realloc(readbuf, bufsize * 2);
		if (!newbuf) {
			if (total_readed < st.st_size) {
				WARN("%s: MEM WARNING: %llu bytes only readed from %llu\n",
					file_name, (LLU)total_readed, (LLU)st.st_size);
			} else {
				WARN("%s: MEM WARNING: %llu bytes only readed.\n",
					file_name, (LLU)total_readed);
			}
			break;
		}
		readbuf = newbuf;
		readpos += readed;
		bufsize *= 2;
	}

	if (!is_stdin)
		fclose(fp_in);

	if (!is_stdin && total_readed < st.st_size && total_readed < bufsize) {
		WARN("%s: READ WARNING: filesize=%llu, readed=%llu, error %d, %s\n",
			file_name, (LLU)st.st_size, (LLU)total_readed, errno, strerror(errno));
	}

	if (getenv("SKEIN_DUMP")) {
		printf("DEBUG: file=%s, bufsize=%llu (0x%llx), total_readed=%llu, buf:\n",
			file_name, (LLU)bufsize, (LLU)bufsize, (LLU)total_readed);
		fwrite(readbuf, 1, total_readed, stdout);
	}

	unsigned char output[hashbitlen/4];
	Hash(hashbitlen, (unsigned char*) readbuf, total_readed, output);

	free(readbuf);

	hash2hexstr(output, MsgDigest);

	return 1;
}

int PrintFileHash(const char filename[], int tag, char mode)
{
	char MsgDigest[hashbitlen/2];
	if (HashFile(filename, MsgDigest, mode) < 0)
		return -1;
	if (tag == 1) {
		printf("skein%d_v%s (%s) = %s\n", hashbitlen, skeinVersion, filename, MsgDigest);
	} else {
		printf("%s %s%s\n", MsgDigest, (mode == 'b' ? "*" : ""), filename);
	}
	return 1;
}

/* Return: -1 = Error, 0 = Mismatch, 1 = Match */
int HashMatch(const char StoredDigest[], const char *filename, int quiet)
{
	char mode = 't';
	if (filename[0] == '*') {
		filename++;
		mode = 'b';
	}
	
	char MsgDigest[hashbitlen/2];
	if (HashFile(filename, MsgDigest, mode) < 0)
		return -1;

	if (strcmp(MsgDigest, StoredDigest)) {
		printf("%s: FAILED\n", filename);
		return 0;
	}

	if (quiet > 0)
		printf("%s: OK\n", filename);
	return 1;
}

int isProper(const char MsgDigest[])
{
	int len = strlen(MsgDigest);
	if (len != (hashbitlen / 4))
		return 0;
	int index = 0;
	for (index = 0; index < len; index++)
	{
		char c = MsgDigest[index];
		if (c >= '0' && c <= '9') continue;
		if (c >= 'A' && c <= 'F') continue;
		return 0;
	}

	return 1;
}


int decomposeHashLine(char hash[], char MsgDigest_tmp[], char file_tmp[])
{
	char c = 0;
	int i = 0 , j =0;
	int isTagFile = 0;
	char alg[20];
	char tmp[1000];
	while(((c = hash[i])!=' ')&&((c = hash[i])!='_'))
	{
		tmp[i] = hash[i];
		i++;
	}
	tmp[i] = 0;

	sprintf(alg,"skein%d",hashbitlen);
	if(!strcmp(alg,tmp))
	{
		isTagFile = 1;
	}

	if(isTagFile == 0)
	{
		strcpy(MsgDigest_tmp,tmp);
		i++;
		while((c = hash[i])!= '\n')
		{
			file_tmp[j] = hash[i];
			i++;
			j++;
		}
		file_tmp[j] = 0;
	}
	else if((hash[i]=='_')&&(hash[i+1]=='v'))
	{
		i = i + 2;
		j = 0;
		char version[5];
		while((c = hash[i])!=' ')
		{
			version[j] = hash[i];
			i++;
			j++;
		}
		version[i] = 0;
		float vers = 0, skeinVers = 0;
		sscanf(version,"%f",&vers);
		sscanf(skeinVersion,"%f",&skeinVers);

		j = 0;
		i = i + 2;
		while((c = hash[i])!=')')
		{
			file_tmp[j] = hash[i];
			i++;
			j++;
		}
		file_tmp[j] = 0;

		i = i + 4;
		j = 0;
		while((c = hash[i])!='\n')
		{
			MsgDigest_tmp[j] = hash[i];
			i++;
			j++;
		}
		MsgDigest_tmp[j] = 0;

		if(skeinVers < vers)
		{//version newer than mine
			return (-1);
		}
		else if (skeinVers > vers)
		{//going to use older version than mine
			return (0) ;
		}
		else
		{ //versions match
			return (1);
		}
	}
	return 1;

}

/* Return: -1 = some errors/mismatches, 1 = all ok */
int VerifyHashesFromFile(FILE *fp, int status, int warn, int quiet)
{
	char hash[PATH_MAX + hashbitlen/4 + 4];
	char MsgDigest_tmp[hashbitlen/2];
	int NoMatch = 0, NotProper = 0, Computed = 0;
	int line = 0;

	while (fgets(hash, sizeof(hash)-1, fp))
	{
		char file_tmp[PATH_MAX];
		line ++;
		Computed++;
		int hashVersion = decomposeHashLine(hash,MsgDigest_tmp,file_tmp);
		if (hashVersion == -1)
		{
			WARN("%s is using newer version of skein%d algorithm\n"
			     "You should update your algorithm\n",
			     file_tmp, hashbitlen);
			continue;
		}
		else if (hashVersion == 0)
		{
			WARN("%s is using an older version of skein%d algorithm\n"
			     "You should use the older algorithm\n",
			     file_tmp, hashbitlen);
			continue;
		}
		else if (!isProper(MsgDigest_tmp))
		{
			if(status != 1 && warn == 1)
				WARN("%s: %d: improperly formatted skein%d checksum line\n",file_tmp,line,hashbitlen);
			NotProper ++;
			NoMatch ++;
		}
		else if (HashMatch(MsgDigest_tmp, file_tmp, quiet) <= 0)
		{
			NoMatch++;
		}
	}
	if(NoMatch)
	{
		WARN("WARNING: %d of %d computed checksums did NOT match\n",
			NoMatch, Computed);
	}
	if(NotProper)
	{
		WARN("WARNING: %d line is improperly formatted\n", NotProper);
	}
	return (NotProper || NoMatch) ? -1 : 1;
}

void PrintStringHash(const char *s)
{
	unsigned char output[hashbitlen/4];
	char digest[hashbitlen/4 + 1];
	Hash(hashbitlen, s, strlen(s), output);
	hash2hexstr(output, digest);
	printf("%s -%s\n", digest, s);
}

void print_version(void)
{
	printf("skein%dsum 1.0\n", hashbitlen);
	printf("License GPLv3+: GNU GPL version 3 or later\n");
	printf("<http://gnu.org/licenses/gpl.html>\n");
	printf("This is free software: you are free to change and redistribute it.\n");
	printf("There is NO WARRANTY, to the extent permitted by law.\n");
	exit(1);
}

void print_usage(void)
{
	printf("Usage: skein%dsum [OPTION]... [FILE]...\n",hashbitlen);
	printf("Print or check skein (%d-bit) checksums.\n",hashbitlen);
	printf("With no FILE, or when FILE is -, read standard input.\n"
	       "\n"
	       "-b, --binary     read in binary mode\n"
	       "-c, --check      read skein sums from the FILEs and check them\n"
	       "--tag            create a BSD-style checksum\n"
	       "-t, --text       read in text mode (default)\n"
	       "-0               hash strings from command line\n"
	       "\n"
	       "The following three options are useful only when verifying checksums:\n"
	       "--quiet          don't print OK for each successfully verified file\n"
	       "--status         don't output anything, status code shows success\n"
	       "-w, --warn       warn about improperly formatted checksum lines\n"
	       "\n"
	       "--strict         with --check, exit non-zero for any invalid input\n"
	       "-h, --help       display this help and exit\n"
	       "-V, --version    output version information and exit\n"
	       "\n"
	       "The sums are computed as described in version 1.3 of the Skein\n"
	       "specification. When checking, the input should be a former output of\n"
	       "this program. The default mode is to print a line with checksum, a\n"
	       "character indicating input mode ('*' for binary, space for text), and\n"
	       "name for each FILE.\n");
	exit(1);
}

int is_goodfile(const char filename[])
{
	if (!strcmp(filename, "-"))
		return 1;

	struct stat s;

	if (stat(filename, &s) < 0) {
		WARN("%s: no such file or directory\n", filename);
		return 0;
	}

	if (S_ISREG (s.st_mode)) return 1;
	if (S_ISCHR (s.st_mode)) return 1;
	if (S_ISBLK (s.st_mode)) return 1;
	if (S_ISLNK (s.st_mode)) return 1;
	if (S_ISFIFO(s.st_mode)) return 1;

	if (S_ISDIR (s.st_mode)) {
		WARN("%s: is a directory\n", filename);
		return 0;
	}

	WARN("%s: WARNING: unknown filetype 0x%Xu\n", filename, s.st_mode);
	return 1;  /* try it as good */
}


int main(int argc, char** argv)
{
	int first_file = argc;
	int binary = -1,
		 check = -1,
		 quiet = -1,
		 warn = -1,
		 status = -1,
		 tag = -1,
		 hashString = -1;

	int errorFound = 0;
	int opt = 0;
/*****************************************************************************************
 ************************************* GETTING DATA ***********************************
 *****************************************************************************************/
	while ((opt = getopt_long (argc, argv, "hVbctw0", long_options, NULL)) != -1)
	{
		switch (opt) {
			case '0'            : hashString = 1;  break;
			case 'b'            : binary     = 1;  break;
			case 't'            : binary     = 0;  break;
			case 'c'            : check      = 1;  break;
			case 'w'            : warn       = 1;  break;
			case QUIET_OPTION   : quiet      = 1;  break;
			case STATUS_OPTION  : status     = 1;  break;
			case TAG_OPTION     : tag        = 1;  break;
			
			case 'h'            : print_usage();   /* ..never returns */
			case 'V'            : print_version(); /* ..never returns */

			default: TRYHELP_GOODBYE();
		}
	}

	first_file = optind;

/*****************************************************************************************
 ************************************* PROCESSING DATA ***********************************
 *****************************************************************************************/

	if (hashString > 0)
	{
		int n = first_file;
		if (n >= argc) {
			WARN("command line should contain strings for hashing\n");
			TRYHELP_GOODBYE();
		}
		while (n < argc)
			PrintStringHash(argv[n++]);
		return 0;
	}

	if (check < 0)   /* READ FILES, GENERATE CHECKSUMS AND PRINT TO STDOUT.. */
	{
		if (quiet == 1 || warn == 1 || status == 1)
		{
			if(quiet == 1)
				WARN("the --quiet option is meaningful only when verifying checksums\n");
			if(status ==1)
				WARN("the --status option is meaningful only when verifying checksums\n");
			if(warn == 1)
				WARN("the --warn option is meaningful only when verifying checksums\n");
			TRYHELP_GOODBYE();
		}

		char mode = binary > 0 ? 'b' : 't';

		int file_index;
		for (file_index = first_file; file_index < argc; file_index++)
		{
			const char *filename = argv[file_index];
			if (!is_goodfile(filename)) {
				errorFound++;
				continue;
			}

			if (binary > 0 && hashString == 1) {
				WARN("%s: No such file or directory\n", filename);
				continue;
			}

			if (PrintFileHash(filename, tag, mode) < 0)
				errorFound++;
		}

		if (first_file >= argc)  /* no files in cmdline? hash stdin.. */
			if (PrintFileHash("-", tag, mode) < 0)
				errorFound++;

		return (errorFound ? 1 : 0);
	}

	if (check == 1)   /* READ LISTFILES, GENERATE HASHES, COMPARE WITH STORED, PRINT RESULT */
	{
		if (tag == 1 || binary >= 0)
		{
			if (tag == 1)
				WARN("the --tag option is meaningless when verifying checksums\n");
			if (binary >= 0)
				WARN("the --text and --binary options are meaningless when verifying checksums\n");
			TRYHELP_GOODBYE();
		}

		int file_index;
		for (file_index = first_file; file_index < argc; file_index++)
		{
			const char *filename = argv[file_index];
			if (!is_goodfile(filename)) {
				errorFound++;
				continue;
			}

			FILE *fp = (strcmp(filename, "-") == 0) ? stdin : fopen(filename, "r");
			if (!fp) {
				errorFound++;
				WARN("%s: %s\n", filename, strerror(errno));
				continue;
			}
			if (VerifyHashesFromFile(fp, status, warn, quiet) < 0)
				errorFound++;
			if (fp != stdin)
				fclose(fp);
		}

		if (first_file >= argc)  /* no files in cmdline? verify stdin.. */
			if (VerifyHashesFromFile(stdin, status, warn, quiet) < 0)
				errorFound++;

		return (errorFound ? 1 : 0);
	}

	return 1;
}
